import io
import logging
import mmap
from dataclasses import dataclass
from functools import singledispatchmethod
from typing import Dict, Optional, Type
from uuid import UUID

import archinfo

from cle.errors import CLEError, CLEUnknownFormatError

from . import Backend, register_backend
from .pe import PE
from .te import TE

try:
    import uefi_firmware
except ImportError:
    uefi_firmware = None

log = logging.getLogger(__name__)


class UefiDriverLoadError(Exception):
    """
    This error is raised (and caught internally) if the data contained in the UEFI entity tree doesn't make sense.
    """


class UefiFirmware(Backend):
    """
    A UEFI firmware blob loader. Support is provided by the ``uefi_firmware`` package.
    """

    is_default = True

    @classmethod
    def _to_bytes(cls, fileobj: io.IOBase):
        try:
            fileno = fileobj.fileno()
        except io.UnsupportedOperation:
            pass
        else:
            return mmap.mmap(fileno, 0, access=mmap.ACCESS_READ)

        if isinstance(fileobj, io.BytesIO):
            return fileobj.getbuffer()

        # fuck it, we'll do it live
        fileobj.seek(0)
        return fileobj.read()

    @classmethod
    def is_compatible(cls, stream):
        if uefi_firmware is None:
            return False

        buffer = cls._to_bytes(stream)
        parser = uefi_firmware.AutoParser(buffer)
        return parser.type() != "unknown"

    def __init__(self, *args, **kwargs) -> None:
        if uefi_firmware is None:
            raise CLEError("run `pip install uefi_firmware==1.10` to load UEFI firmware")
        super().__init__(*args, **kwargs)

        # hack: we are using a loader internal method in a non-kosher way which will cause our children to be
        # marked as the main binary if we are also the main binary
        # work around this by setting ourself here:
        if self.loader._main_object is None:
            self.loader._main_object = self

        self._drivers: Dict[UUID, "UefiModuleMixin"] = {}
        self._drivers_pending: Dict[UUID, "UefiModulePending"] = {}
        self._current_file: Optional[UUID] = None

        self.set_arch(archinfo.arch_from_id("x86_64"))  # TODO: ???

        buffer = self._to_bytes(self._binary_stream)
        parser = uefi_firmware.AutoParser(buffer)
        firmware = parser.parse()
        self._load(firmware)

        while self._drivers_pending:
            uuid, pending = self._drivers_pending.popitem()
            try:
                child = pending.build(self, uuid)
            except UefiDriverLoadError as e:
                log.warning("Failed to load %s: %s", uuid, e.args[0])
            else:
                self._drivers[uuid] = child
                child.parent_object = self
                self.child_objects.append(child)

        if self.child_objects:
            self._arch = self.child_objects[0].arch
        else:
            log.warning("Loaded empty UEFI firmware?")
        self.has_memory = False
        self.pic = True

        # hack pt. 2
        if self.loader._main_object is self:
            self.loader._main_object = None

    @singledispatchmethod
    def _load(self, uefi_obj):  # pylint: disable=no-self-use
        raise CLEUnknownFormatError(f"Can't load firmware object: {uefi_obj}")

    if uefi_firmware is not None:

        @_load.register
        def _load_generic(self, uefi_obj: uefi_firmware.FirmwareObject):
            for obj in uefi_obj.objects:
                self._load(obj)

        @_load.register
        def _load_none(self, uefi_obj: None):
            pass

        @_load.register
        def _load_firmwarefile(self, uefi_obj: uefi_firmware.uefi.FirmwareFile):
            old_uuid = self._current_file
            if uefi_obj.type == 7:  # driver
                uuid = UUID(bytes=uefi_obj.guid)
                self._drivers_pending[uuid] = UefiModulePending()
                self._current_file = uuid
            self._load_generic(uefi_obj)
            self._current_file = old_uuid

        @_load.register
        def _load_firmwarefilesection(self, uefi_obj: uefi_firmware.uefi.FirmwareFileSystemSection):
            pending = self._drivers_pending.get(self._current_file, None)
            if pending is not None:
                if uefi_obj.type == 16:  # pe32 image
                    pending.pe_image = uefi_obj.content
                elif uefi_obj.type == 18:  # te image
                    pending.te_image = uefi_obj.content
                elif uefi_obj.type == 21:  # user interface name
                    pending.name = uefi_obj.content.decode("utf-16").strip("\0")
            self._load_generic(uefi_obj)


@dataclass
class UefiModulePending:
    """
    A worklist entry for the UEFI firmware loader.
    """

    name: Optional[str] = None
    pe_image: Optional[bytes] = None
    te_image: Optional[bytes] = None
    # version
    # dependencies

    def build(self, parent: UefiFirmware, guid: UUID) -> "UefiModuleMixin":
        count = (self.pe_image is not None) + (self.te_image is not None)
        if count > 1:
            raise UefiDriverLoadError("Multiple image sections")
        cls: "Type[UefiModuleMixin]"
        if self.pe_image is not None:
            cls = UefiPE
            data = self.pe_image
        elif self.te_image is not None:
            cls = UefiTE
            data = self.te_image
        else:
            raise UefiDriverLoadError("Missing PE or TE image section")
        return cls(None, io.BytesIO(data), is_main_bin=False, loader=parent.loader, name=self.name, guid=guid)


class UefiModuleMixin(Backend):
    """
    A mixin to make other kinds of backends load as UEFI modules.
    """

    def __init__(self, *args, guid: UUID, name: Optional[str], **kwargs):
        super().__init__(*args, **kwargs)
        self.guid = guid
        self.user_interface_name = name

        if self.linked_base == 0:
            self.pic = True

    def __repr__(self):
        return (
            f"<{type(self).__name__} Object "
            f'{self.guid}{f" {self.user_interface_name}" if self.user_interface_name else ""}, '
            f"maps [{self.min_addr:#x}:{self.max_addr:#x}]>"
        )


class UefiPE(UefiModuleMixin, PE):
    """
    A PE file contained in a UEFI image.
    """


class UefiTE(UefiModuleMixin, TE):
    """
    A TE file contained in a UEFI image.
    """


register_backend("uefi", UefiFirmware)



import logging

from .elfreloc import ELFReloc
from .generic import (
    GenericAbsoluteAddendReloc,
    GenericCopyReloc,
    GenericIRelativeReloc,
    GenericJumpslotReloc,
    GenericRelativeReloc,
    GenericTLSDescriptorReloc,
    GenericTLSDoffsetReloc,
    GenericTLSModIdReloc,
    GenericTLSOffsetReloc,
)

log = logging.getLogger(name=__name__)

# http://infocenter.arm.com/help/topic/com.arm.doc.ihi0056b/IHI0056B_aaelf64.pdf
arch = "AARCH64"


class R_AARCH64_ABS64(GenericAbsoluteAddendReloc):
    pass


class R_AARCH64_COPY(GenericCopyReloc):
    pass


class R_AARCH64_GLOB_DAT(GenericJumpslotReloc):
    pass


class R_AARCH64_JUMP_SLOT(GenericJumpslotReloc):
    pass


class R_AARCH64_RELATIVE(GenericRelativeReloc):
    pass


class R_AARCH64_IRELATIVE(GenericIRelativeReloc):
    pass


class R_AARCH64_TLS_DTPREL(GenericTLSDoffsetReloc):
    pass


class R_AARCH64_TLS_DTPMOD(GenericTLSModIdReloc):
    pass


class R_AARCH64_TLS_TPREL(GenericTLSOffsetReloc):
    pass


class R_AARCH64_TLSDESC(GenericTLSDescriptorReloc):
    RESOLVER_ADDR = 0xFFFF_FFFF_FFFF_FE00


class R_AARCH64_CALL26(ELFReloc):
    """
    Relocation Type: 283
    Calculation: (S + A - P)
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        P = self.rebased_addr
        return S + A - P

    def relocate(self):
        if not self.resolved:
            return False
        if not ((-(2**27)) <= self.value and self.value < (2**27)):
            log.warning("relocation out of range")
        instr = self.owner.memory.unpack_word(self.relative_addr, size=4) & 0b11111100000000000000000000000000
        imm = self.value >> 2 & 0x3FFFFFF
        self.owner.memory.pack_word(self.relative_addr, instr | imm, size=4)
        return True


class R_AARCH64_ADR_PREL_PG_HI21(ELFReloc):
    """
    Relocation Type: 275
    Calculation: Page(S + A) - Page(P)
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        P = self.rebased_addr
        return ((S + A) & ~0xFFF) - (P & ~0xFFF)

    def relocate(self):
        if not self.resolved:
            return False
        if not ((-(2**32)) <= self.value and self.value < (2**32)):
            log.warning("relocation out of range")
        instr = self.owner.memory.unpack_word(self.relative_addr, size=4) & 0b10011111000000000000000000011111
        imm = self.value >> 12 & 0x1FFFFF
        immlo = imm & 0b11
        immhi = imm >> 2
        self.owner.memory.pack_word(self.relative_addr, instr | (immhi << 5) | (immlo << 29), size=4)
        return True


class R_AARCH64_ADD_ABS_LO12_NC(ELFReloc):
    """
    Relocation Type: 275
    Calculation: (S + A)
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        return S + A

    def relocate(self):
        if not self.resolved:
            return False
        instr = self.owner.memory.unpack_word(self.relative_addr, size=4) & 0b11111111110000000000001111111111
        imm = self.value & 0xFFF
        self.owner.memory.pack_word(self.relative_addr, instr | (imm << 10), size=4)
        return True



import importlib
import logging
import os
from collections import defaultdict

import archinfo

from cle.backends.relocation import Relocation

ALL_RELOCATIONS = defaultdict(dict)
complaint_log = set()

path = os.path.dirname(os.path.abspath(__file__))
log = logging.getLogger(name=__name__)


def load_relocations():
    for filename in os.listdir(path):
        if not filename.endswith(".py"):
            continue
        if filename == "__init__.py":
            continue

        log.debug("Importing PE relocation module: %s", filename[:-3])
        module = importlib.import_module(".%s" % filename[:-3], "cle.backends.pe.relocation")

        try:
            arch_name = module.arch
        except AttributeError:
            continue

        for item_name in dir(module):
            if item_name not in archinfo.defines:
                continue
            item = getattr(module, item_name)
            if not isinstance(item, type) or not issubclass(item, Relocation):
                continue

            ALL_RELOCATIONS[arch_name][archinfo.defines[item_name]] = item


def get_relocation(arch, r_type):
    if r_type == 0:
        return None
    try:
        return ALL_RELOCATIONS[arch][r_type]
    except KeyError:
        if (arch, r_type) not in complaint_log:
            complaint_log.add((arch, r_type))
            log.warning("Unknown reloc %d on %s", r_type, arch)
        return None


load_relocations()



from .generic import (
    GenericAbsoluteAddendReloc,
    GenericCopyReloc,
    GenericIRelativeReloc,
    GenericJumpslotReloc,
    GenericPCRelativeAddendReloc,
    GenericRelativeReloc,
    GenericTLSDoffsetReloc,
    GenericTLSModIdReloc,
    GenericTLSOffsetReloc,
)

arch = "X86"


class R_386_32(GenericAbsoluteAddendReloc):
    pass


class R_386_PC32(GenericPCRelativeAddendReloc):
    pass


class R_386_COPY(GenericCopyReloc):
    pass


class R_386_GLOB_DAT(GenericJumpslotReloc):
    pass


class R_386_JMP_SLOT(GenericJumpslotReloc):
    pass


class R_386_RELATIVE(GenericRelativeReloc):
    pass


class R_386_IRELATIVE(GenericIRelativeReloc):
    pass


class R_386_TLS_DTPMOD32(GenericTLSModIdReloc):
    pass


class R_386_TLS_TPOFF(GenericTLSOffsetReloc):
    pass


class R_386_TLS_DTPOFF32(GenericTLSDoffsetReloc):
    pass



from .generic import (
    IMAGE_REL_BASED_DIR64,
    IMAGE_REL_BASED_HIGH,
    IMAGE_REL_BASED_HIGHADJ,
    IMAGE_REL_BASED_HIGHLOW,
    IMAGE_REL_BASED_LOW,
)
from .pereloc import PEReloc

arch = "arm"


class IMAGE_REL_BASED_HIGHADJ(IMAGE_REL_BASED_HIGHADJ):
    pass


class IMAGE_REL_BASED_DIR64(IMAGE_REL_BASED_DIR64):
    pass


class IMAGE_REL_BASED_HIGHLOW(IMAGE_REL_BASED_HIGHLOW):
    pass


class IMAGE_REL_BASED_HIGH(IMAGE_REL_BASED_HIGH):
    pass


class IMAGE_REL_BASED_LOW(IMAGE_REL_BASED_LOW):
    pass


class IMAGE_REL_BASED_ARM_MOV32(PEReloc):
    pass


class IMAGE_REL_BASED_THUMB_MOV32(PEReloc):
    pass



from .generic import (
    GenericAbsoluteAddendReloc,
    GenericCopyReloc,
    GenericRelativeReloc,
    GenericTLSDoffsetReloc,
    GenericTLSModIdReloc,
    GenericTLSOffsetReloc,
)

arch = "MIPS64"


class R_MIPS_64(GenericAbsoluteAddendReloc):
    pass


class R_MIPS_REL32(GenericRelativeReloc):
    pass


class R_MIPS_COPY(GenericCopyReloc):
    pass


class R_MIPS_TLS_DTPMOD64(GenericTLSModIdReloc):
    pass


class R_MIPS_TLS_DTPREL64(GenericTLSDoffsetReloc):
    pass


class R_MIPS_TLS_TPREL64(GenericTLSOffsetReloc):
    pass



from .generic import (
    GenericAbsoluteAddendReloc,
    GenericCopyReloc,
    GenericIRelativeReloc,
    GenericJumpslotReloc,
    GenericPCRelativeAddendReloc,
    GenericRelativeReloc,
    GenericTLSDoffsetReloc,
    GenericTLSModIdReloc,
    GenericTLSOffsetReloc,
    RelocGOTMixin,
    RelocTruncate32Mixin,
)

arch = "AMD64"


class R_X86_64_64(GenericAbsoluteAddendReloc):
    pass


class R_X86_64_COPY(GenericCopyReloc):
    pass


class R_X86_64_RELATIVE(GenericRelativeReloc):
    pass


class R_X86_64_IRELATIVE(GenericIRelativeReloc):
    pass


class R_X86_64_GLOB_DAT(GenericJumpslotReloc):
    pass


class R_X86_64_JUMP_SLOT(GenericJumpslotReloc):
    pass


class R_X86_64_DTPMOD64(GenericTLSModIdReloc):
    pass


class R_X86_64_DTPOFF64(GenericTLSDoffsetReloc):
    pass


class R_X86_64_TPOFF64(GenericTLSOffsetReloc):
    pass


class R_X86_64_PC32(RelocTruncate32Mixin, GenericPCRelativeAddendReloc):
    check_sign_extend = True


class R_X86_64_32(RelocTruncate32Mixin, GenericAbsoluteAddendReloc):
    check_zero_extend = True


class R_X86_64_32S(RelocTruncate32Mixin, GenericAbsoluteAddendReloc):
    check_sign_extend = True


class R_X86_64_PLT32(RelocTruncate32Mixin, GenericPCRelativeAddendReloc):
    check_sign_extend = True


class R_X86_64_GOTPCREL(RelocGOTMixin, RelocTruncate32Mixin, GenericPCRelativeAddendReloc):
    check_sign_extend = True


class R_X86_64_GOTPCRELX(RelocGOTMixin, RelocTruncate32Mixin, GenericPCRelativeAddendReloc):
    check_sign_extend = True


class R_X86_64_REX_GOTPCRELX(RelocGOTMixin, RelocTruncate32Mixin, GenericPCRelativeAddendReloc):
    check_sign_extend = True



from .generic import (
    GenericAbsoluteAddendReloc,
    GenericCopyReloc,
    GenericIRelativeReloc,
    GenericJumpslotReloc,
    GenericRelativeReloc,
    GenericTLSOffsetReloc,
)

arch = "S390X"


class R_390_GLOB_DAT(GenericJumpslotReloc):
    pass


class R_390_JMP_SLOT(GenericJumpslotReloc):
    pass


class R_390_RELATIVE(GenericRelativeReloc):
    pass


class R_390_64(GenericAbsoluteAddendReloc):
    pass


class R_390_TLS_TPOFF(GenericTLSOffsetReloc):
    pass


class R_390_IRELATIVE(GenericIRelativeReloc):
    pass


class R_390_COPY(GenericCopyReloc):
    pass



from .arm import (
    R_ARM_ABS32,
    R_ARM_ABS32_NOI,
    R_ARM_CALL,
    R_ARM_COPY,
    R_ARM_GLOB_DAT,
    R_ARM_JUMP24,
    R_ARM_JUMP_SLOT,
    R_ARM_MOVT_ABS,
    R_ARM_MOVW_ABS_NC,
    R_ARM_PC24,
    R_ARM_PREL31,
    R_ARM_REL32,
    R_ARM_REL32_NOI,
    R_ARM_RELATIVE,
    R_ARM_THM_CALL,
    R_ARM_THM_JUMP6,
    R_ARM_THM_JUMP19,
    R_ARM_THM_JUMP24,
    R_ARM_THM_MOVT_ABS,
    R_ARM_THM_MOVW_ABS_NC,
    R_ARM_TLS_DTPOFF32,
    R_ARM_TLS_TPOFF32,
)

arch = "ARMCortexM"

__all__ = [
    "arch",
    "R_ARM_CALL",
    "R_ARM_PREL31",
    "R_ARM_REL32",
    "R_ARM_ABS32",
    "R_ARM_MOVW_ABS_NC",
    "R_ARM_MOVT_ABS",
    "R_ARM_THM_CALL",
    "R_ARM_COPY",
    "R_ARM_GLOB_DAT",
    "R_ARM_JUMP_SLOT",
    "R_ARM_RELATIVE",
    "R_ARM_ABS32_NOI",
    "R_ARM_REL32_NOI",
    "R_ARM_TLS_DTPOFF32",
    "R_ARM_TLS_TPOFF32",
    "R_ARM_JUMP24",
    "R_ARM_PC24",
    "R_ARM_THM_JUMP24",
    "R_ARM_THM_JUMP19",
    "R_ARM_THM_JUMP6",
    "R_ARM_THM_MOVW_ABS_NC",
    "R_ARM_THM_MOVT_ABS",
]



from .generic import (
    GenericAbsoluteAddendReloc,
    GenericAbsoluteReloc,
    GenericRelativeReloc,
    GenericTLSDoffsetReloc,
    GenericTLSModIdReloc,
    GenericTLSOffsetReloc,
)

arch = "MIPS32"


class R_MIPS_32(GenericAbsoluteAddendReloc):
    pass


class R_MIPS_REL32(GenericRelativeReloc):
    pass


class R_MIPS_JUMP_SLOT(GenericAbsoluteReloc):
    pass


class R_MIPS_GLOB_DAT(GenericAbsoluteReloc):
    pass


class R_MIPS_TLS_DTPMOD32(GenericTLSModIdReloc):
    pass


class R_MIPS_TLS_TPREL32(GenericTLSOffsetReloc):
    pass


class R_MIPS_TLS_DTPREL32(GenericTLSDoffsetReloc):
    pass


class R_MIPS_HI16(GenericAbsoluteReloc):
    def relocate(self):
        if not self.resolved:
            return False

        self.owner.memory.pack_word(self.dest_addr, self.value >> 16, size=2)
        return True


class R_MIPS_LO16(GenericAbsoluteReloc):
    def relocate(self):
        if not self.resolved:
            return False

        self.owner.memory.pack_word(self.dest_addr, self.value & 0xFFFF, size=2)
        return True
